@@ -7,7 +7,7 @@ use Carp ();
 use IO::Handle ();
 use Exporter 'import';
 
-our $VERSION = '0.009';
+our $VERSION = '0.010';
 
 use constant DEFAULT_REQUEST_BODY_LIMIT => 16777216;
 use constant DEFAULT_REQUEST_BODY_BUFFER => 262144;
@@ -306,9 +306,7 @@ sub _body_params {
       foreach my $part (@{$self->_body_multipart}) {
         next if defined $part->{filename};
         my ($name, $value, $headers) = @$part{'name','content','headers'};
-        if (uc $default_charset eq 'UTF-8' and do { local $@; eval { require Unicode::UTF8; 1 } }) {
-          $name = Unicode::UTF8::decode_utf8($name);
-        } elsif (length $default_charset) {
+        if (length $default_charset) {
           require Encode;
           $name = Encode::decode($default_charset, "$name");
         }
@@ -320,9 +318,7 @@ sub _body_params {
               $value_charset = defined $charset_quoted ? $charset_quoted : $charset_unquoted;
             }
           }
-          if (uc $value_charset eq 'UTF-8' and do { local $@; eval { require Unicode::UTF8; 1 } }) {
-            $value = Unicode::UTF8::decode_utf8($value);
-          } elsif (length $value_charset) {
+          if (length $value_charset) {
             require Encode;
             $value = Encode::decode($value_charset, "$value");
           }
@@ -368,10 +364,7 @@ sub _body_uploads {
       foreach my $part (@{$self->_body_multipart}) {
         next unless defined $part->{filename};
         my ($name, $filename, $file, $size, $headers) = @$part{'name','filename','file','size','headers'};
-        if (uc $default_charset eq 'UTF-8' and do { local $@; eval { require Unicode::UTF8; 1 } }) {
-          $name = Unicode::UTF8::decode_utf8($name);
-          $filename = Unicode::UTF8::decode_utf8($filename);
-        } elsif (length $default_charset) {
+        if (length $default_charset) {
           require Encode;
           $name = Encode::decode($default_charset, "$name");
           $filename = Encode::decode($default_charset, "$filename");
@@ -425,11 +418,13 @@ sub _body_multipart {
       $input = defined $self->{input_handle} ? $self->{input_handle} : *STDIN;
       binmode $input;
     }
-    my $parts = _parse_multipart($input, $length, $boundary, $self->{request_body_buffer});
+
+    my $parts = _parse_multipart($input, $length, $boundary, $self->{request_body_buffer} || $ENV{CGI_TINY_REQUEST_BODY_BUFFER});
     unless (defined $parts) {
       $self->{response_status} = "400 $HTTP_STATUS{400}" unless $self->{headers_rendered};
       die "Malformed multipart/form-data request\n";
     }
+
     $self->{body_parts} = $parts;
   }
   return $self->{body_parts};
@@ -452,22 +447,37 @@ sub set_response_status {
   if ($self->{headers_rendered}) {
     Carp::carp "Attempted to set HTTP response status but headers have already been rendered";
   } else {
-    if ($status =~ m/\A[0-9]+ [^\r\n]*\z/) {
+    if (defined $status and $status =~ m/\A[0-9]+ [^\r\n]*\z/) {
       $self->{response_status} = $status;
-    } else {
+    } elsif (defined $status) {
       Carp::croak "Attempted to set unknown HTTP response status $status" unless exists $HTTP_STATUS{$status};
       $self->{response_status} = "$status $HTTP_STATUS{$status}";
+    } else {
+      delete $self->{response_status};
     }
   }
   return $self;
 }
 
+sub set_response_content_disposition {
+  my ($self, $disposition, $filename) = @_;
+  if ($self->{headers_rendered}) {
+    Carp::carp "Attempted to set HTTP response content disposition but headers have already been rendered";
+  } else {
+    Carp::croak "Invalid characters in HTTP response content disposition" if defined $disposition and $disposition =~ m/[^a-zA-Z0-9!#\$%&'*+\-.^_`|~]/;
+    $self->{response_content_disposition} = $disposition;
+    # filename will be quoted/escaped later
+    $self->{response_filename} = $filename;
+  }
+  return $self;
+}
+
 sub set_response_content_type {
   my ($self, $content_type) = @_;
   if ($self->{headers_rendered}) {
     Carp::carp "Attempted to set HTTP response content type but headers have already been rendered";
   } else {
-    Carp::croak "Newline characters not allowed in HTTP response content type" if $content_type =~ tr/\r\n//;
+    Carp::croak "Newline characters not allowed in HTTP response content type" if defined $content_type and $content_type =~ tr/\r\n//;
     $self->{response_content_type} = $content_type;
   }
   return $self;
@@ -475,22 +485,11 @@ sub set_response_content_type {
 
 sub set_response_charset {
   my ($self, $charset) = @_;
-  Carp::croak "Invalid characters in HTTP response charset" if $charset =~ m/[^a-zA-Z0-9!#\$%&'*+\-.^_`|~]/;
+  Carp::croak "Invalid characters in HTTP response charset" if defined $charset and $charset =~ m/[^a-zA-Z0-9!#\$%&'*+\-.^_`|~]/;
   $self->{response_charset} = $charset;
   return $self;
 }
 
-sub set_response_download {
-  my ($self, $filename) = @_;
-  if ($self->{headers_rendered}) {
-    Carp::carp "Attempted to set HTTP response content disposition but headers have already been rendered";
-  } else {
-    $self->{response_attachment} = 1;
-    $self->{response_filename} = $filename;
-  }
-  return $self;
-}
-
 sub add_response_header {
   my ($self, $name, $value) = @_;
   if ($self->{headers_rendered}) {
@@ -531,6 +530,8 @@ sub add_response_header {
   }
 }
 
+sub reset_response_headers { delete $_[0]{response_headers}; $_[0] }
+
 sub response_status_code {
   my ($self) = @_;
   if (defined $self->{response_status} and $self->{response_status} =~ m/\A([0-9]+)/) {
@@ -559,6 +560,20 @@ sub headers_rendered { $_[0]{headers_rendered} }
         $headers_str .= "$name: $value\r\n";
         $headers_set{lc $name} = 1;
       }
+      if (!$headers_set{'content-disposition'} and (defined $self->{response_content_disposition} or defined $self->{response_filename})) {
+        my $value = defined $self->{response_content_disposition} ? $self->{response_content_disposition} : 'inline';
+        if (defined(my $filename = $self->{response_filename})) {
+          require Encode;
+          my $quoted_filename = Encode::encode('ISO-8859-1', "$filename");
+          $quoted_filename =~ tr/\r\n/  /;
+          $quoted_filename =~ s/([\\"])/\\$1/g;
+          $value .= "; filename=\"$quoted_filename\"";
+          my $ext_filename = Encode::encode('UTF-8', "$filename");
+          $ext_filename =~ s/([^a-zA-Z0-9!#\$&+\-.^_`|~])/sprintf '%%%02X', ord $1/ge;
+          $value .= "; filename*=UTF-8''$ext_filename";
+        }
+        $headers_str = "Content-Disposition: $value\r\n$headers_str" unless lc $value eq 'inline';
+      }
       if (!$headers_set{location} and $type eq 'redirect') {
         Carp::croak "Newline characters not allowed in HTTP redirect" if $data =~ tr/\r\n//;
         $headers_str = "Location: $data\r\n$headers_str";
@@ -574,22 +589,6 @@ sub headers_rendered { $_[0]{headers_rendered} }
           unless defined $content_type;
         $headers_str = "Content-Type: $content_type\r\n$headers_str";
       }
-      if (!$headers_set{'content-disposition'} and $self->{response_attachment}) {
-        my $filename = $self->{response_filename};
-        my $value = 'attachment';
-        if (defined $filename and length $filename) {
-          require Encode;
-          my $quoted_filename = Encode::encode('ISO-8859-1', $filename);
-          $quoted_filename =~ tr/\r\n/  /;
-          $quoted_filename =~ s/([\\"])/\\$1/g;
-          $value .= "; filename=\"$quoted_filename\"";
-          my $ext_filename = $filename;
-          utf8::encode $ext_filename;
-          $ext_filename =~ s/([^a-zA-Z0-9!#\$&+\-.^_`|~])/sprintf '%%%02X', ord $1/ge;
-          $value .= "; filename*=UTF-8''$ext_filename";
-        }
-        $headers_str .= "Content-Disposition: $value\r\n";
-      }
       if (!$headers_set{date}) {
         my $date_str = epoch_to_date(time);
         $headers_str = "Date: $date_str\r\n$headers_str";
@@ -665,7 +664,7 @@ sub _parse_multipart {
       $buffer .= $$input;
       $length = 0;
     } else {
-      my $chunk = $buffer_size || $ENV{CGI_TINY_REQUEST_BODY_BUFFER} || DEFAULT_REQUEST_BODY_BUFFER;
+      my $chunk = $buffer_size || DEFAULT_REQUEST_BODY_BUFFER;
       $chunk = $length if $length < $chunk;
       last unless my $read = read $input, $buffer, $chunk, length $buffer;
       $length -= $read;
@@ -703,13 +702,11 @@ sub _parse_multipart {
 
           $state{part}{headers}{lc $name} = $value;
           if (lc $name eq 'content-disposition') {
-            if (my ($name_quoted, $name_unquoted) = $value =~ m/;\s*name\s*=\s*(?:"((?:\\[\\"]|[^"])*)"|([^";]*))/i) {
-              $name_quoted =~ s/\\([\\"])/$1/g if defined $name_quoted;
-              $state{part}{name} = defined $name_quoted ? $name_quoted : $name_unquoted;
-            }
-            if (my ($filename_quoted, $filename_unquoted) = $value =~ m/;\s*filename\s*=\s*(?:"((?:\\[\\"]|[^"])*)"|([^";]*))/i) {
-              $filename_quoted =~ s/\\([\\"])/$1/g if defined $filename_quoted;
-              $state{part}{filename} = defined $filename_quoted ? $filename_quoted : $filename_unquoted;
+            while ($value =~ m/;\s*([^=\s]+)\s*=\s*(?:"((?:\\[\\"]|[^"])*)"|([^";]*))/ig) {
+              my ($field_name, $field_quoted, $field_unquoted) = ($1, $2, $3);
+              next unless lc $field_name eq 'name' or lc $field_name eq 'filename';
+              $field_quoted =~ s/\\([\\"])/$1/g if defined $field_quoted;
+              $state{part}{lc $field_name} = defined $field_quoted ? $field_quoted : $field_unquoted;
             }
           }
         }
